# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, List, Dict
from datetime import datetime
import pickle
import re
from sqlalchemy import create_engine, Column, Unicode, Float, LargeBinary
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session, Session
from . import JobStore, JobDoesNotExistException, JobAlreadyExistsException
from ..job import Job


Base = declarative_base()
SessionCls = scoped_session(sessionmaker())


def db_user(func):
    def wrapper(self, *args, **kwargs):
        session = kwargs.get('session', None)
        session_exists = True

        if session is None:
            session = SessionCls()
            kwargs['session'] = session
            session_exists = False

        result = func(self, *args, **kwargs)

        if not session_exists:
            session.commit()
            session.close()

        return result
    return wrapper


class DbJob(Base):
    __tablename__ = 'yapsched_jobs'

    id = Column(Unicode(191, _warn_on_bytestring=False), primary_key=True)
    next_run_time = Column(Float(25), index=True)
    job_state = Column(LargeBinary, nullable=False)


class SqlAlchemyJobStore(JobStore):
    def __init__(self, url: str = None, engine: Engine = None, engine_options: Optional[Dict] = None,
                 pickle_protocol=pickle.HIGHEST_PROTOCOL):
        super().__init__()
        self.pickle_protocol = pickle_protocol

        if engine is not None:
            self.engine: Engine = engine
        elif url is not None:
            self.engine: Engine = create_engine(url, **(engine_options or {}))
        else:
            raise ValueError('Need either "engine" or "url" defined')

        SessionCls.configure(bind=self.engine)

    def setup(self):
        super().setup()
        Base.metadata.create_all(self.engine)

    def teardown(self):
        super().teardown()
        self.engine.dispose()

    @db_user
    def add_job(self, job: Job, replace_existing: bool = True, session: Session = None):
        if self.contains_job(job.id):
            if not replace_existing:
                raise JobAlreadyExistsException(job.id)

            self.update_job(job, session=session)
            return

        if job.next_run_time is None and job.active:
            job.next_run_time = job.trigger.get_next_fire_time(None, None)

        self._logger.debug(f'Add job ({job}) next run time: {job.next_run_time}')

        new_job = DbJob(id=job.id,
                        next_run_time=job.get_next_run_time_ts(),
                        job_state=pickle.dumps(job.__getstate__(), self.pickle_protocol))
        session.add(new_job)

    @db_user
    def update_job(self, job: Job, session: Session = None):
        existing_job = self._get_db_job(job.id, session)

        if job.next_run_time is None and job.active:
            if existing_job.next_run_time is None:
                job.next_run_time = job.trigger.get_next_fire_time(None, None)
            else:
                job.next_run_time = datetime.fromtimestamp(existing_job.next_run_time, tz=self._scheduler.tz)

        self._logger.debug(f'Update job ({job}) next run time: {job.next_run_time}')

        existing_job.next_run_time = job.get_next_run_time_ts()
        existing_job.job_state = pickle.dumps(job.__getstate__(), self.pickle_protocol)

    @db_user
    def remove_job(self, job_id: str, session: Session = None):
        existing_job = self._get_db_job(job_id, session)
        self._logger.debug(f'Delete job ({self._reconstitute_job(existing_job)})')
        session.delete(existing_job)

    @db_user
    def remove_all_jobs(self, session=None):
        session.query(DbJob).delete()

    @db_user
    def get_job(self, job_id: str, session: Session = None) -> Job:
        existing_job = self._get_db_job(job_id, session)
        return self._reconstitute_job(existing_job)

    @db_user
    def get_jobs(self, pattern: str = None, session: Session = None) -> List[Job]:
        db_jobs = session.query(DbJob).order_by(DbJob.next_run_time.asc()).all()

        if pattern is not None:
            compiled_pattern = re.compile(pattern)
            db_jobs = filter(lambda db_job: compiled_pattern.fullmatch(db_job.id), db_jobs)

        return list(map(lambda db_job: self._reconstitute_job(db_job), db_jobs))

    @db_user
    def contains_job(self, job_id: str, session: Session = None) -> bool:
        try:
            self._get_db_job(job_id, session)
            return True
        except JobDoesNotExistException:
            return False

    @staticmethod
    def _get_db_job(job_id: str, session: Session) -> DbJob:
        existing_job = session.query(DbJob).get(job_id)
        if existing_job is None:
            raise JobDoesNotExistException(job_id)
        return existing_job

    @staticmethod
    def _reconstitute_job(db_job: DbJob):
        job_state = pickle.loads(db_job.job_state)
        job = object.__new__(Job)
        job.__setstate__(job_state)
        return job



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing
from queue import Empty
from threading import Thread
from typing import Optional, Dict, List
from datetime import datetime
from collections import defaultdict
import random
import logging
import contextlib
from .job import Job
from . import events

random.seed()
logger = logging.getLogger(__name__)


class Executor:
    _lock: Optional[multiprocessing.RLock] = None

    def __init__(self, pool_size: int, scheduler):
        self._pool_size = pool_size  # TODO: use _pool_size
        self._scheduler = scheduler

        self._instances: Dict[int, Dict] = {}
        self._jobs_to_instances: Dict[str, List[int]] = defaultdict(list)
        self._recent_instance_ids: List[int] = []

        self.mp_ctx = multiprocessing.get_context('spawn')

        self._event_queue: Optional[multiprocessing.Queue] = None
        self._event_thread: Optional[Thread] = None
        self._running = False

    def setup(self):
        self._lock = self.mp_ctx.RLock()
        self._event_queue = self.mp_ctx.Queue()

        self._event_thread = Thread(target=self._event_read_loop, daemon=True)
        self._running = True
        self._event_thread.start()

    def teardown(self, wait: bool = True):
        logger.debug('Tearing down...')

        processes = [instance['process'] for instance in self._instances.values()]
        logger.debug(f'Processes: {processes}')

        def get_alive_procs():
            procs = []
            for proc in processes:
                try:
                    if proc.is_alive():
                        procs.append(proc)
                except ValueError:
                    pass
            return procs

        while alive_processes := get_alive_procs():
            logger.debug(f'Alive processes: {alive_processes}')
            for process in alive_processes:
                logger.debug(f'Process: {process}')
                if not wait:
                    logger.debug('    terminating')
                    process.terminate()
                logger.debug('    joining')
                process.join(1)
                if not process.is_alive():
                    logger.debug('    closing')
                    try:
                        process.close()
                    except ValueError:
                        pass
                else:
                    logger.debug('    still alive; will try again soon')

        self._instances.clear()
        self._jobs_to_instances.clear()

        self._lock = None

        logger.debug('Closing event thread')
        self._running = False
        self._event_thread.join()
        logger.debug('Deleting event thread')
        del self._event_thread
        logger.debug('Closing event queue')
        self._event_queue.close()
        self._event_queue.join_thread()

        logger.debug('Done')

    def submit_job(self, job: Job, run_time: datetime) -> int:
        if self._lock is None:
            raise ExecutorNotSetupException

        with self._lock:
            if job.max_instances != -1 and len(self._jobs_to_instances[job.id]) >= job.max_instances:
                raise MaxJobInstancesReachedException(job, self._jobs_to_instances[job.id])

            dynamic_args = {}
            for arg_name, values in job.dynamic_args.items():
                func = values['func']
                dynamic_args[arg_name] = func()

            instance_id = self._gen_new_instance_id()
            process = self.mp_ctx.Process(target=run_job, name=f'Process-{instance_id}',
                                          args=(job, instance_id, run_time, self._event_queue),
                                          kwargs={**job.kwargs, **dynamic_args})

            self._instances[instance_id] = {
                'job': job,
                'id': instance_id,
                'start_time': datetime.now(tz=self._scheduler.tz),
                'process': process
            }

            self._jobs_to_instances[job.id].append(instance_id)

            try:
                logger.info(f'Running job "{job.id}" (scheduled at {run_time})')

                with contextlib.redirect_stdout(None), contextlib.redirect_stderr(None):
                    process.start()
            except Exception:
                self._job_cleanup(job.id, instance_id, terminate=True)
                raise

            return instance_id

    def terminate_job(self, instance_id: int) -> str:
        with self._lock:
            instance = self._instances.get(instance_id, None)

            if instance is None:
                raise JobInstanceNotFoundException(instance_id)

        job_id = instance['job'].id
        self._job_cleanup(job_id, instance_id, terminate=True)

        return job_id

    def get_job_instance_ids(self, job_id: str) -> List[int]:
        with self._lock:
            return self._jobs_to_instances.get(job_id, [])

    def get_job(self, instance_id: int) -> Job:
        with self._lock:
            instance = self._instances.get(instance_id, None)
            if instance is None:
                raise JobInstanceNotFoundException(instance_id)

            return instance['job']

    def get_instances(self) -> List[dict]:
        with self._lock:
            return list(self._instances.values())

    def _event_read_loop(self):
        logger.debug('waiting for event...')
        while self._running:
            try:
                event: events.JobExecutionEvent = self._event_queue.get(timeout=0.2)
            except Empty:
                continue

            job = self._instances[event.instance_id]['job']

            logger.debug(f'got event ({event}) from job {job.id}:{event.instance_id}; '
                         f'event queue size: {self._event_queue.qsize()}')

            if event.success:
                logger.info(f'Job "{job.id}" executed successfully')
                if job.callback:
                    try:
                        job.callback(job.id, event.instance_id, event.retval)
                    except Exception as e:
                        logger.error(f'callback failed!')
                        logger.exception(e)
            else:
                logger.error(f'Job "{job.id}" raised an exception\n{event.formatted_exc}')
                if job.error_callback:
                    try:
                        job.error_callback(job.id, event.instance_id, event.exc)
                    except Exception as e:
                        logger.error(f'error callback failed!')
                        logger.exception(e)

            self._job_cleanup(event.job_id, event.instance_id)
            self._scheduler._dispatch_event(event)

            logger.debug('waiting for event...')

    def _job_cleanup(self, job_id: str, instance_id: int, terminate: bool = False):
        with self._lock:
            process = self._instances[instance_id]['process']

            logger.debug(f'cleaning up instance (job_id: {job_id}, instance_id: {instance_id}, process: {process})')

            if process.is_alive():
                if terminate:
                    logger.debug(f'terminating process')
                    try:
                        process.terminate()
                    except Exception as e:
                        logger.exception(e)
                process.join(timeout=2)

            if not process.is_alive():
                process.close()
                logger.debug('process closed')
            else:
                logger.error('Could not terminate process')

            logger.debug(f'instances: {self._instances}')
            logger.debug(f'mapping: {dict(self._jobs_to_instances)}')

            logger.debug('removing instance')
            del self._instances[instance_id]
            self._jobs_to_instances[job_id].remove(instance_id)

            logger.debug(f'instances: {self._instances}')
            logger.debug(f'mapping: {dict(self._jobs_to_instances)}')

            logger.debug('cleaned up')

    def _gen_new_instance_id(self) -> int:
        all_possible_ids = set(range(10000, 100000))

        with self._lock:
            current_ids = set(self._instances.keys()) | set(self._recent_instance_ids)
            available_ids = list(all_possible_ids - current_ids)

            instance_id = random.choice(available_ids)

            self._recent_instance_ids.append(instance_id)
            if len(self._recent_instance_ids) > 10000:
                self._recent_instance_ids.pop(0)

            return instance_id


def run_job(job: Job, instance_id: int, run_time: datetime, event_queue: multiprocessing.Queue, **kwargs):
    try:
        retval = job.func(job.id, instance_id, *job.args, **job.kwargs, **kwargs)
        logger.debug(f'job {job.id}:{instance_id} returned {retval}')
    except Exception as e:
        logger.debug(f'job {job.id}:{instance_id} raised an exception')
        event = events.JobExecutionEvent(events.EVENT_JOB_EXECUTED, job.id, instance_id, run_time, success=False,
                                         exc=e)
    else:
        logger.debug(f'job {job.id}:{instance_id} executed successfully')
        event = events.JobExecutionEvent(events.EVENT_JOB_EXECUTED, job.id, instance_id, run_time, success=True,
                                         retval=retval)

    logger.debug(f'sending event to queue ({job.id}:{instance_id})')
    [h.flush() for h in logger.handlers]
    event_queue.put(event)
    logger.debug(f'event queue size: {event_queue.qsize()} ({job.id}:{instance_id})')


class ExecutorNotSetupException(Exception):
    def __init__(self):
        super().__init__('Executor not setup')


class MaxJobInstancesReachedException(Exception):
    def __init__(self, job: Job, instance_ids: List[int]):
        super().__init__(f'Job "{job.id}" has already reached its maximum number of instances '
                         f'({job.max_instances})')
        self.job = job
        self.instance_ids = instance_ids


class JobInstanceNotFoundException(Exception):
    def __init__(self, instance_id: int):
        super().__init__(f'Job instance {instance_id} not found')
        self.instance_id = instance_id



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional, List
from ..util import get_class_logger
from ..job import Job


class JobStore(ABC):
    def __init__(self):
        self._logger = get_class_logger(self)
        self._scheduler = None

    def setup(self):
        pass

    def teardown(self):
        pass

    def get_due_jobs(self, latest: datetime) -> List[Job]:
        jobs = self.get_jobs()
        pending = list(filter(lambda job: job.active and job.next_run_time <= latest, jobs))
        return pending

    def get_next_run_time(self) -> Optional[datetime]:
        jobs = self.get_jobs()
        return jobs[0].next_run_time if jobs else None

    @abstractmethod
    def add_job(self, job: Job, replace_existing: bool):
        pass

    @abstractmethod
    def update_job(self, job: Job):
        pass

    @abstractmethod
    def remove_job(self, job_id: str):
        pass

    @abstractmethod
    def remove_all_jobs(self):
        pass

    @abstractmethod
    def get_job(self, job_id: str) -> Job:
        pass

    @abstractmethod
    def get_jobs(self, pattern: str = None) -> List[Job]:
        pass

    @abstractmethod
    def contains_job(self, job_id: str) -> bool:
        pass


class JobAlreadyExistsException(Exception):
    def __init__(self, job_id):
        super().__init__(f'Job "{job_id}" already exists')


class JobDoesNotExistException(Exception):
    def __init__(self, job_id):
        super().__init__(f'Job "{job_id}" does not exist')



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from datetime import datetime
from typing import Any
import traceback

EVENT_SCHEDULER_STARTED = 2 ** 0
EVENT_SCHEDULER_STOPPED = 2 ** 1
EVENT_SCHEDULER_PAUSED = 2 ** 2
EVENT_SCHEDULER_RESUMED = 2 ** 3
EVENT_ALL_JOBS_REMOVED = 2 ** 4
EVENT_JOB_ADDED = 2 ** 5
EVENT_JOB_REMOVED = 2 ** 6
EVENT_JOB_MODIFIED = 2 ** 7
EVENT_JOB_EXECUTED = 2 ** 8
EVENT_JOB_ERROR = 2 ** 9
EVENT_JOB_MISSED = 2 ** 10
EVENT_JOB_SUBMITTED = 2 ** 11
EVENT_JOB_MAX_INSTANCES = 2 ** 12
EVENT_JOB_TERMINATED = 2 ** 13

# UPDATE THIS WHENEVER AN EVENT IS ADDED/REMOVED!
# if the last event is 2 ** x, this should be 2 ** (x + 1) - 1
EVENT_ALL = 2 ** 14 - 1


@dataclass
class Event:
    code: int


class SchedulerEvent(Event):
    pass


@dataclass
class JobEvent(Event):
    job_id: str


class JobExecutionEvent(JobEvent):
    def __init__(self, code: int, job_id: str, instance_id: int, run_time: datetime, success: bool, retval: Any = None,
                 exc: Exception = None):
        super().__init__(code, job_id)

        self.instance_id = instance_id
        self.run_time = run_time
        self.success = success
        self.retval = retval
        self.exc = exc

        if self.exc is not None:
            self.formatted_exc = ''.join(traceback.format_exception(type(exc), exc, exc.__traceback__)).rstrip()
        else:
            self.formatted_exc = None



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, auto
from typing import Optional, Callable, List, Dict
import threading
import uuid
from datetime import datetime, timezone, timedelta
import logging
from .job import Job
from . import executor, jobstores, triggers, events

logger = logging.getLogger(__name__)


class State(Enum):
    STOPPED = auto()
    RUNNING = auto()
    PAUSED = auto()


class Scheduler:
    _event: Optional[threading.Event] = None
    _thread: Optional[threading.Thread] = None

    def __init__(self, pool_size: int = 10, jobstore: Optional[jobstores.JobStore] = None, tz=timezone.utc):
        if tz is None:
            raise ValueError('tz must not be None')

        self._executor = executor.Executor(pool_size, self)
        self._executor_lock = threading.RLock()

        self._jobstore = None
        self._jobstore_lock = threading.RLock()

        self._listeners = []
        self._listeners_lock = threading.RLock()

        self.state = State.STOPPED
        self._instances = {}

        self.tz = None

        self.configure(jobstore=jobstore, tz=tz)

    def configure(self, jobstore: Optional[jobstores.JobStore] = None, tz=None):
        if jobstore is not None:
            jobstore._scheduler = self
        self._jobstore = jobstore

        if tz is not None:
            self.tz = tz

    def start(self, paused: bool = False):
        if self.state != State.STOPPED:
            raise SchedulerAlreadyRunningException

        self._event = threading.Event()

        with self._executor_lock:
            self._executor.setup()

        with self._jobstore_lock:
            self._jobstore.setup()

        self.add_listener(self._internal_listener)

        self.state = State.PAUSED if paused else State.RUNNING
        logger.info('Scheduler started')
        self._dispatch_event(events.SchedulerEvent(events.EVENT_SCHEDULER_STARTED))

        if not paused:
            self.wakeup()

        self._thread = threading.Thread(target=self._main_loop, name='yapsched', daemon=True)
        self._thread.start()

    def shutdown(self, wait: bool = True):
        if self.state == State.STOPPED:
            raise SchedulerNotRunningException

        logger.info('Shutting down...')

        self.state = State.STOPPED

        logger.debug('Closing thread')
        self._event.set()
        self._thread.join()

        logger.debug('Deleting thread')
        del self._thread

        logger.debug('Tearing down executor')
        with self._executor_lock:
            self._executor.teardown(wait)

        logger.debug('Tearing down jobstore')
        with self._jobstore_lock:
            self._jobstore.teardown()

        self._dispatch_event(events.SchedulerEvent(events.EVENT_SCHEDULER_STOPPED))

        self.remove_listener(self._internal_listener)

        logger.info('Scheduler shutdown')

    def pause(self):
        if self.state == State.STOPPED:
            raise SchedulerNotRunningException
        elif self.state == State.RUNNING:
            self.state = State.PAUSED
            logger.info('Scheduler paused')
            self._dispatch_event(events.SchedulerEvent(events.EVENT_SCHEDULER_PAUSED))

    def resume(self):
        if self.state == State.STOPPED:
            raise SchedulerNotRunningException
        elif self.state == State.PAUSED:
            self.state = State.RUNNING
            logger.info('Scheduler resumed')
            self._dispatch_event(events.SchedulerEvent(events.EVENT_SCHEDULER_RESUMED))
            self.wakeup()

    @property
    def running(self) -> bool:
        return self.state != State.STOPPED

    def wakeup(self):
        self._event.set()

    def add_job(self,
                func: Callable,
                args: Optional[List] = None,
                kwargs: Optional[Dict] = None,
                dynamic_args: Optional[Dict[str, Callable]] = None,
                identifier: Optional[str] = None,
                description: Optional[str] = None,
                coalesce: bool = False,
                max_instances: int = -1,
                replace_existing: bool = False,
                trigger: Optional[triggers.Trigger] = None,
                next_run_time: Optional[datetime] = None,
                active: bool = True,
                callback: Optional[Callable] = None,
                error_callback: Optional[Callable] = None) -> Job:
        if next_run_time is not None and not active:
            raise ValueError('Job can\'t be inactive and have a non-null next_run_time!')

        if identifier is None:
            identifier = uuid.uuid4()

        if trigger is None:
            trigger = triggers.DateTrigger(datetime.now(tz=self.tz))

        trigger.tz = self.tz

        job = Job(func,
                  identifier,
                  tuple(args) if args is not None else (),
                  kwargs if kwargs is not None else {},
                  dynamic_args if dynamic_args is not None else {},
                  trigger,
                  next_run_time=next_run_time,
                  description=description,
                  coalesce=coalesce,
                  max_instances=max_instances,
                  active=active,
                  callback=callback,
                  error_callback=error_callback)

        with self._jobstore_lock:
            self._jobstore.add_job(job, replace_existing)

        logger.info(f'Added job {identifier}')
        self._dispatch_event(events.JobEvent(events.EVENT_JOB_ADDED, identifier))

        if self.state == State.RUNNING and active:
            self.wakeup()

        return job

    def modify_job(self, job_id: str, **changes) -> Job:
        with self._jobstore_lock:
            job = self.get_job(job_id)

            new_next_run_time = changes['next_run_time'] if 'next_run_time' in changes else job.next_run_time
            new_active = changes['active'] if 'active' in changes else job.active

            if new_next_run_time is not None and not new_active:
                raise ValueError('Job can\'t be inactive and have a non-null next_run_time!')

            job._modify(**changes)

            if job.trigger is None:
                job.trigger = triggers.DateTrigger(datetime.now(tz=self.tz))

            job.trigger.tz = self.tz

            self._jobstore.update_job(job)

        self._dispatch_event(events.JobEvent(events.EVENT_JOB_MODIFIED, job_id))

        if self.state == State.RUNNING:
            self.wakeup()

        return job

    def get_job(self, job_id: str) -> Job:
        with self._jobstore_lock:
            return self._jobstore.get_job(job_id)

    def get_jobs(self, pattern: str = None) -> List[Job]:
        with self._jobstore_lock:
            return self._jobstore.get_jobs(pattern)

    def remove_job(self, job_id: str):
        with self._jobstore_lock:
            self._jobstore.remove_job(job_id)

        logger.info(f'Removed job {job_id}')
        self._dispatch_event(events.JobEvent(events.EVENT_JOB_REMOVED, job_id))

    def terminate_job(self, instance_id: int):
        with self._executor_lock:
            job_id = self._executor.terminate_job(instance_id)
            logger.info(f'Terminated job {job_id}')
            self._dispatch_event(events.JobEvent(events.EVENT_JOB_TERMINATED, job_id))

    def trigger_job(self, job_id: str) -> int:
        job = self.get_job(job_id)
        run_time = datetime.now(tz=self.tz)

        try:
            instance_id = self._executor.submit_job(job, run_time)
        except executor.MaxJobInstancesReachedException:
            self._dispatch_event(events.JobEvent(events.EVENT_JOB_MAX_INSTANCES, job_id))
            raise

        return instance_id

    def get_job_instance_ids(self, job_id: str) -> List[int]:
        with self._jobstore_lock:
            if not self._jobstore.contains_job(job_id):
                raise jobstores.JobDoesNotExistException(job_id)

        with self._executor_lock:
            return self._executor.get_job_instance_ids(job_id)

    def get_job_with_instance_id(self, instance_id: int) -> Job:
        with self._executor_lock:
            return self._executor.get_job(instance_id)

    def get_running_jobs(self) -> List[dict]:
        with self._executor_lock:
            return self._executor.get_instances()

    def add_listener(self, callback: Callable, mask: int = events.EVENT_ALL):
        with self._listeners_lock:
            self._listeners.append((callback, mask))

    def remove_listener(self, callback):
        with self._listeners_lock:
            for i, (cb, _) in enumerate(self._listeners):
                if callback == cb:
                    del self._listeners[i]

    def _main_loop(self):
        max_wait = threading.TIMEOUT_MAX
        wait_seconds = max_wait
        while self.state != State.STOPPED:
            self._event.wait(wait_seconds)
            self._event.clear()
            if self.state == State.STOPPED:
                break
            wait_seconds = self._process_jobs()
            if wait_seconds == -1:
                wait_seconds = max_wait

    def _process_jobs(self) -> int:
        if self.state == State.PAUSED:
            logger.debug('Scheduler paused; not processing jobs')
            return -1

        logger.debug('Processing jobs')
        now_dt = datetime.now(tz=self.tz)
        next_wakeup_time: Optional[datetime] = None
        events_to_dispatch = []

        with self._jobstore_lock:
            try:
                due_jobs = self._jobstore.get_due_jobs(now_dt)
                logger.debug(f'Due jobs: {due_jobs}')

                for job in due_jobs:
                    run_times = job._get_run_times(now_dt)
                    run_times = run_times[-1:] if job.coalesce else run_times
                    logger.debug(f'Job: {job}; run times: {run_times}')

                    for run_time in run_times:
                        try:
                            self._executor.submit_job(job, run_time)
                        except executor.MaxJobInstancesReachedException as e:
                            logger.error(f'Execution of job skipped: {e}')
                            events_to_dispatch.append(events.JobEvent(events.EVENT_JOB_MAX_INSTANCES, job.id))
                        except Exception as e:
                            logger.exception(e)
                        else:
                            events_to_dispatch.append(events.JobEvent(events.EVENT_JOB_SUBMITTED, job.id))

                    job_next_run_time = job.trigger.get_next_fire_time(run_times[-1], None)
                    logger.debug(f'Job next run time: {job_next_run_time}')
                    if job_next_run_time:
                        job.next_run_time = job_next_run_time
                        self._jobstore.update_job(job)
                    else:
                        self.remove_job(job.id)
            except Exception as e:
                logger.warning(f'Error getting due jobs from job store: {e}')
                next_wakeup_time = now_dt + timedelta(seconds=5)

            jobstore_next_run_time = self._jobstore.get_next_run_time()
            logger.debug(f'Jobstore next run time: {jobstore_next_run_time}')
            if jobstore_next_run_time is not None and next_wakeup_time is None:
                next_wakeup_time = jobstore_next_run_time

        for event in events_to_dispatch:
            self._dispatch_event(event)

        if self.state == State.PAUSED:
            wait_seconds = -1
            logger.debug('Scheduler is paused; waiting for resume')
        elif next_wakeup_time is None:
            wait_seconds = -1
            logger.debug('No jobs; waiting until a job is added')
        else:
            wait_seconds = min(max((next_wakeup_time - now_dt).total_seconds(), 0), int(threading.TIMEOUT_MAX))
            logger.debug(f'Next wakeup is due at {next_wakeup_time} (in {wait_seconds} seconds)')

        return wait_seconds

    def _dispatch_event(self, event: events.Event):
        with self._listeners_lock:
            listeners = tuple(self._listeners)

        for callback, mask in listeners:
            if event.code & mask:
                try:
                    callback(event)
                except Exception as e:
                    logger.error('Error notifying listener')
                    logger.exception(e)

    def _internal_listener(self, event: events.Event):
        if isinstance(event, events.JobExecutionEvent):
            if event.code == events.EVENT_JOB_EXECUTED:
                with self._jobstore_lock:
                    logger.info(f'Next wakeup time: {self._jobstore.get_next_run_time()}')


class SchedulerAlreadyRunningException(Exception):
    def __init__(self):
        super().__init__('Scheduler is already running')


class SchedulerNotRunningException(Exception):
    def __init__(self):
        super().__init__('Scheduler is not running')


class JobAlreadyExistsException(Exception):
    def __init__(self, job_id: str):
        super().__init__(f'Job already exists with ID {job_id}')
        self.job_id = job_id



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Callable, Dict, Tuple, List
from datetime import datetime
import inspect
from .triggers import Trigger
from . import util


class Job:
    """
    A job to be executed by the executor.
    """
    def __init__(self,
                 func: Callable,
                 identifier: str,
                 args: Tuple,
                 kwargs: Dict,
                 dynamic_args: Dict[str, Callable],
                 trigger: Trigger,
                 next_run_time: Optional[datetime] = None,
                 description: Optional[str] = None,
                 coalesce: bool = False,
                 max_instances: int = -1,
                 active: bool = True,
                 callback: Optional[Callable] = None,
                 error_callback: Optional[Callable] = None):
        """
        Initialize the job.
        :param func: the function to be called when the job executes
        :param identifier: the unique ID of the job
        :param args: the arguments to pass to func
        :param kwargs: the keyword arguments to pass to func
        :param dynamic_args: a dictionary of argument names mapping to functions that will be executed upon job
            execution. the return values of the functions will be passed to func as keyword arguments
        :param trigger: the trigger to used to determine when the job will execute
        :param next_run_time: the next run time of the job. if None, the next run time will be determined by the trigger
        :param description: the description of the job
        :param coalesce: if True, will only run once when several run times are due
        :param max_instances: the maximum number of concurrently running instances of the job. -1 means no limit, 0
            means disabled
        :param callback: a function to call if func executes successfully. it must accept three arguments: job_id,
            instance_id, and retval. the retval argument will be the return value of func
        :param error_callback: a function to call if func raises an exception. it must accept three arguments: job_id,
            instance_id, and exc. the exc argument will be the exception instance
        """
        self.func = func
        self.id = identifier
        self.args = args
        self.kwargs = kwargs
        self.trigger = trigger
        self.next_run_time = next_run_time
        self.description = description
        self.coalesce = coalesce
        self.max_instances = max_instances
        self.active = active

        try:
            self.func_ref = util.obj_to_ref(self.func)
        except ValueError:
            self.func_ref = None

        if callback is None:
            self.callback = dummy_callback
        else:
            self.callback = callback

        if error_callback is None:
            self.error_callback = dummy_callback
        else:
            self.error_callback = error_callback

        try:
            self.callback_ref = util.obj_to_ref(self.callback)
        except ValueError:
            self.callback_ref = None

        try:
            self.error_callback_ref = util.obj_to_ref(self.error_callback)
        except ValueError:
            self.error_callback_ref = None

        self.dynamic_args = {}
        for arg_name, arg_func in dynamic_args.items():
            try:
                arg_func_ref = util.obj_to_ref(arg_func)
            except ValueError:
                arg_func_ref = None

            self.dynamic_args[arg_name] = {'func': arg_func, 'func_ref': arg_func_ref}

    def _modify(self, **changes):
        pass

    def _get_run_times(self, latest: datetime) -> List[datetime]:
        run_times = []
        next_run_time = self.next_run_time
        while next_run_time and next_run_time < latest:
            run_times.append(next_run_time)
            next_run_time = self.trigger.get_next_fire_time(next_run_time, latest)
        return run_times

    def get_next_run_time_ts(self):
        return self.next_run_time.timestamp() if self.next_run_time is not None else None

    def __getstate__(self):
        if self.func_ref is None:
            raise Exception(f'Job function ({self.func}) not serializable')

        if self.callback_ref is None:
            raise Exception(f'Job callback function ({self.callback}) not serializable')

        if self.error_callback_ref is None:
            raise Exception(f'Job error callback function ({self.error_callback}) not serializable')

        dynamic_args = {}
        for arg_name, values in self.dynamic_args.items():
            func = values['func']
            func_ref = values['func_ref']
            if func_ref is None:
                raise Exception(f'Job dynamic arg function ({func}) not serializable')
            dynamic_args[arg_name] = func_ref

        if inspect.ismethod(self.func) and not inspect.isclass(self.func.__self__):
            args = (self.func.__self__,) + tuple(self.args)
        else:
            args = self.args

        return {
            'id': self.id,
            'func': self.func_ref,
            'args': args,
            'kwargs': self.kwargs,
            'dynamic_args': dynamic_args,
            'trigger': self.trigger,
            'next_run_time': self.next_run_time,
            'description': self.description,
            'coalesce': self.coalesce,
            'max_instances': self.max_instances,
            'active': self.active,
            'callback': self.callback_ref,
            'error_callback': self.error_callback_ref,
        }

    def __setstate__(self, state):
        self.id = state['id']

        self.func_ref = state['func']
        self.func = util.ref_to_obj(self.func_ref)

        self.args = state['args']
        self.kwargs = state['kwargs']
        self.trigger = state['trigger']
        self.next_run_time = state['next_run_time']
        self.description = state['description']
        self.coalesce = state['coalesce']
        self.max_instances = state['max_instances']
        self.active = state['active']

        self.callback_ref = state['callback']
        self.callback = util.ref_to_obj(self.callback_ref)

        self.error_callback_ref = state['error_callback']
        self.error_callback = util.ref_to_obj(self.error_callback_ref)

        self.dynamic_args = {}
        for arg_name, func_ref in state['dynamic_args'].items():
            self.dynamic_args[arg_name] = {'func': util.ref_to_obj(func_ref), 'func_ref': func_ref}

    def __eq__(self, other):
        if isinstance(other, Job):
            return self.id == other.id
        return NotImplemented

    def __repr__(self):
        return f'<Job (id={self.id})>'

    def __str__(self):
        status = f'next run at: {self.next_run_time}' if self.next_run_time is not None else 'INACTIVE'
        return f'{self.id} (trigger: {self.trigger}, {status})'


def dummy_callback(x, y, z):
    pass



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timezone
from abc import ABC, abstractmethod
from typing import Optional
from croniter import croniter
from .util import get_class_logger


class Trigger(ABC):
    def __init__(self):
        self.logger = get_class_logger(self)
        self.tz = timezone.utc

    @abstractmethod
    def get_next_fire_time(self, previous: Optional[datetime] = None, latest: Optional[datetime] = None) -> Optional[datetime]:
        pass

    @abstractmethod
    def __repr__(self):
        pass

    @abstractmethod
    def __str__(self):
        pass


class DateTrigger(Trigger):
    def __init__(self, run_time: datetime):
        super().__init__()
        self.run_time = run_time

    def get_next_fire_time(self, previous: Optional[datetime] = None, latest: Optional[datetime] = None) -> Optional[datetime]:
        next_fire_time = self.run_time if previous is None or self.run_time > previous else None
        return next_fire_time

    def __getstate__(self):
        return {
            'run_time': self.run_time
        }

    def __setstate__(self, state):
        self.run_time = state['run_time']

    def __repr__(self):
        return f'<{self.__class__.__name__} (run_time="{self.run_time}")>'

    def __str__(self):
        return f'date[{self.run_time}]'


class CronTrigger(Trigger):
    def __init__(self, cron_str: str):
        super().__init__()
        self.cron_str = cron_str

    def get_next_fire_time(self, previous: Optional[datetime] = None, latest: Optional[datetime] = None) -> Optional[datetime]:
        if previous is None:
            previous = datetime.now(tz=self.tz)

        itr = croniter(self.cron_str, previous)

        next_fire_time = datetime.fromtimestamp(itr.get_next(), tz=self.tz)

        if latest is not None and next_fire_time > latest:
            return None

        return next_fire_time

    def __repr__(self):
        return f'<{self.__class__.__name__} (cron_str="{self.cron_str}")>'

    def __str__(self):
        return f'cron[{self.cron_str}]'
